// Unofficial VAMS-0808 RS232 interface wrapper
// Originally written by Simon Smith (opensource-ks@killswtch.net)
// http://www.killswtch.net/

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.IO.Ports;
using System.Diagnostics;
using System.Threading;

namespace HALibs.VAMS0808 {
	/// <summary>
	/// Represents a connection to the VAMS0808. There can only be one Connection object per device.
	/// </summary>
	public class Connection {
		public delegate void MessageEventHandler(Connection context, string message);
		public delegate void ChannelMessageEventHandler(Connection context, int channel);
		public delegate void SwitchModeMessageEventHandler(Connection context, SwitchMode mode);

		private Thread messageReaderThread;
		
		private SerialPort _port;
		private MatrixState _state;

		public SerialPort Port {
			get { return _port; }
			private set { _port = value; }
		}

		public MatrixState State {
			get { return _state; }
			private set { _state = value; }
		}

		public bool IsRunning {
			get { return messageReaderThread.IsAlive; }
		}

		public event MessageEventHandler MessageReceived;

		public event ChannelMessageEventHandler InputChannelSelectMessageReceived;
		public event ChannelMessageEventHandler OutputChannelSelectMessageReceived;
		public event MessageEventHandler AllOutputChannelsSelectedMessageReceived;
		public event MessageEventHandler OutputChannelOffMessageReceived;
		public event MessageEventHandler LockEnabledMessageReceived;
		public event MessageEventHandler LockDisabledMessageReceived;
		public event SwitchModeMessageEventHandler SwitchModeMessageReceived;
		public event MessageEventHandler StateMessageReceived;

		public event MessageEventHandler OkMessageReceived;
		public event MessageEventHandler ErrorMessageReceived;
		public event MessageEventHandler NoActionMessageReceived;

		public Connection(SerialPort port) {
			Port = port;
			State = new MatrixState(this);

			Port.NewLine = "\r\n";
			MessageReceived += new MessageEventHandler(Connection_MessageReceived);
		}

		public void Start() {
			Port.Open();

			messageReaderThread = new Thread(new ThreadStart(Run));
			messageReaderThread.Start();
		}

		private void Run() {
			while (Port.IsOpen) {
				StringBuilder currentMessage = new StringBuilder();

				currentMessage.Append((char)Port.ReadChar());

				while (Port.IsOpen) {
					if (Regex.IsMatch(currentMessage.ToString(), "(\r\n|\r|\n)$"))
						break;

					currentMessage.Append((char)Port.ReadChar());
				}

				String currentMessageString = currentMessage.ToString().Trim('\r', '\n', ' ', '\t');

				if (currentMessageString.Length > 0 && MessageReceived != null)
					MessageReceived(this, currentMessageString);
			}
		}

		public void Stop() {
			messageReaderThread.Abort();

			Port.Close();
		}

		void Connection_MessageReceived(Connection context, string message) {
			Trace.WriteLine(String.Format("<-- {0}", message));

			if (Regex.IsMatch(message, "^(0R...?|0ILO.)$")) {
				if (StateMessageReceived != null)
					StateMessageReceived(this, message);
			} else if (Regex.IsMatch(message, "^0CO[1-8]$")) {
				if (OutputChannelSelectMessageReceived != null)
					OutputChannelSelectMessageReceived(this, int.Parse(message.Substring(3, 1)));
			} else if (Regex.IsMatch(message, "^0CI[1-8]$")) {
				if (InputChannelSelectMessageReceived != null)
					InputChannelSelectMessageReceived(this, int.Parse(message.Substring(3, 1)));
			} else if (message == "0ALL") {
				if (AllOutputChannelsSelectedMessageReceived != null)
					AllOutputChannelsSelectedMessageReceived(this, message);
			} else if (message == "0OFF") {
				if (OutputChannelOffMessageReceived != null)
					OutputChannelOffMessageReceived(this, message);
			} else if (message == "0LOI") {
				if (LockEnabledMessageReceived != null)
					LockEnabledMessageReceived(this, message);
			} else if (message == "0LOO") {
				if (LockDisabledMessageReceived != null)
					LockDisabledMessageReceived(this, message);
			} else if (Regex.IsMatch(message, "^0(AV|AC|VC)S$")) {
				if (SwitchModeMessageReceived != null) {
					SwitchMode mode = SwitchMode.AudioVideo;
					if (message == "0ACS")
						mode = SwitchMode.Audio;
					else if (message == "0VCS")
						mode = SwitchMode.Video;

					SwitchModeMessageReceived(this, mode);
				}
			} else if (message == "0GO") {
				if (OkMessageReceived != null)
					OkMessageReceived(this, message);
			} else if (message == "0GN") {
				if (NoActionMessageReceived != null)
					NoActionMessageReceived(this, message);
			} else if (Regex.IsMatch(message, "0E.+")) {
				if (ErrorMessageReceived != null)
					ErrorMessageReceived(this, message);
			} else {
				Debug.WriteLine(String.Format("Unknown message received: {0}", message));
			}
		}

		public bool SelectOutputChannel(int channel) {
			if (SendCommand(String.Format("0CO{0}\r\n", channel), true)) {
				State.CurrentChannelIndex = channel - 1;
				return true;
			}
			return false;
		}

		public bool SelectAllOutputChannels() {
			if (SendCommand("0ALL\r\n", true)) {
				State.SelectAllOutputChannels();
				return true;
			}
			return false;
		}

		public bool SelectInputChannel(int channel) {
			if (SendCommand(String.Format("0CI{0}\r\n", channel), true)) {
				State.ChangeInputChannel(channel);
				return true;
			}
			return false;
		}

		public bool DisableCurrentChannel() {
			//HACK: The VAMS-0808 returns 0GN when using the 0OFF command on all output channels, even if the command was successful
			if (State.AllChannelsSelected) {
				if (SendCommand("0OFF\r\n", "0GN")) {
					State.ChangeInputChannel(0);
					return true;
				}
			} else {
				if (SendCommand("0OFF\r\n", true)) {
					State.ChangeInputChannel(0);
					return true;
				}
			}
			return false;
		}

		public bool RequestStateUpdate() {
			return SendCommand("0RAD\r\n", ".{5,}");
		}

		public bool SetSwitchMode(SwitchMode mode) {
			bool result = false;
			if (mode == SwitchMode.Audio)
				result = SendCommand("0ACS\r\n", true);
			else if (mode == SwitchMode.Video)
				result = SendCommand("0VCS\r\n", true);
			else
				result = SendCommand("0AVS\r\n", true);

			if (result)
				State.Mode = mode;

			return result;
		}

		public bool Lock() {
			if (!State.Locked && SendCommand("0LOI\r\n", true)) {
				State.Locked = true;
				return true;
			}
			return false;
		}

		public bool UnLock() {
			if (State.Locked && SendCommand("0LOO\r\n", true)) {
				State.Locked = false;
				return true;
			}
			return false;
		}

		private bool SendCommand(string command, bool async) {
			if (async) {
				return SendCommand(command, "0GO");
			} else {
				Trace.WriteLine(String.Format("--> {0}", command));
				Port.Write(command);
				return true;
			}
		}

		private bool SendCommand(string command, string okMessageRegex) {
			bool result = true;

			bool done = false;
			result = false;

			MessageEventHandler okHandler = new MessageEventHandler(delegate(Connection context, string message) {
				if (Regex.IsMatch(message, okMessageRegex)) {
					result = true;
				}
				done = true;
			});

			MessageReceived += okHandler;

			for (int i = 0; i < 10 && !done; i++) {
				Trace.WriteLine(String.Format("--> {0}", command));
				Port.Write(command);
				Thread.Sleep(500);
			}

			MessageReceived -= okHandler;

			return result;
		}
	}
}
