// Unofficial VAMS-0808 RS232 interface wrapper
// Originally written by Simon Smith (opensource-ks@killswtch.net)
// http://www.killswtch.net/

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace HALibs.VAMS0808 {
	/// <summary>
	/// Represents the current state of the VAMS-0808.
	/// </summary>
	public class MatrixState {
		public delegate void SwitchModeModifiedEventHandler(SwitchMode previousValue, SwitchMode newValue);
		public delegate void BooleanFieldModifiedEventHandler(bool previousValue, bool newValue);
		public delegate void IntegerFieldModifiedEventHandler(int previousValue, int newValue);

		private ChannelCollection _audioChannels = new ChannelCollection();
		private ChannelCollection _videoChannels = new ChannelCollection();
		private SwitchMode _mode;
		private bool _locked;
		private bool _allChannelsSelected;
		private int _currentChannelIndex;

		public event SwitchModeModifiedEventHandler SwitchModeChanged;
		public event BooleanFieldModifiedEventHandler LockedChanged;
		public event BooleanFieldModifiedEventHandler AllChannelsSelectedChanged;
		public event IntegerFieldModifiedEventHandler CurrentChannelIndexChanged;

		public ChannelCollection AudioChannels {
			get { return _audioChannels; }
		}

		public ChannelCollection VideoChannels {
			get { return _videoChannels; }
		}

		public SwitchMode Mode {
			get { return _mode; }
			internal set {
				SwitchMode oldValue = _mode;
				_mode = value;

				if (oldValue != value && SwitchModeChanged != null)
					SwitchModeChanged(oldValue, value);
			}
		}

		public bool Locked {
			get { return _locked; }
			internal set {
				bool oldValue = _locked;
				_locked = value;

				if (oldValue != value && LockedChanged != null)
					LockedChanged(oldValue, value);
			}
		}

		public bool AllChannelsSelected {
			get { return _allChannelsSelected; }
			protected set {
				bool oldValue = _allChannelsSelected;
				_allChannelsSelected = value;

				if (oldValue != value && AllChannelsSelectedChanged != null)
					AllChannelsSelectedChanged(oldValue, value);
			}
		}

		public int CurrentChannelIndex {
			get { return _currentChannelIndex; }
			internal set {
				int oldValue = _currentChannelIndex;
				_currentChannelIndex = value;

				if (oldValue != value && CurrentChannelIndexChanged != null)
					CurrentChannelIndexChanged(oldValue, value);
			}
		}

		public MatrixState(Connection connection) {
			connection.MessageReceived += new Connection.MessageEventHandler(connection_MessageReceived);

			connection.AllOutputChannelsSelectedMessageReceived += new Connection.MessageEventHandler(connection_AllOutputChannelsSelectedMessageReceived);
			connection.InputChannelSelectMessageReceived += new Connection.ChannelMessageEventHandler(connection_InputChannelSelectMessageReceived);
			connection.OutputChannelSelectMessageReceived += new Connection.ChannelMessageEventHandler(connection_OutputChannelSelectMessageReceived);
			connection.LockDisabledMessageReceived += new Connection.MessageEventHandler(connection_LockDisabledMessageReceived);
			connection.LockEnabledMessageReceived += new Connection.MessageEventHandler(connection_LockEnabledMessageReceived);
			connection.OutputChannelOffMessageReceived += new Connection.MessageEventHandler(connection_OutputChannelOffMessageReceived);
			connection.StateMessageReceived += new Connection.MessageEventHandler(connection_StateMessageReceived);
			connection.SwitchModeMessageReceived += new Connection.SwitchModeMessageEventHandler(connection_SwitchModeMessageReceived);
		}

		void connection_SwitchModeMessageReceived(Connection context, SwitchMode mode) {
			Mode = mode;
			Trace.WriteLine(String.Format("Mode changed to: {0}", mode));
		}

		void connection_StateMessageReceived(Connection context, string message) {
			Trace.WriteLine(String.Format("Updating state using string: {0}", message));
			Update(message);
			Trace.WriteLine("State updated");
		}

		void connection_OutputChannelOffMessageReceived(Connection context, string message) {
			connection_InputChannelSelectMessageReceived(context, 0);
		}

		void connection_LockEnabledMessageReceived(Connection context, string message) {
			Locked = true;
			Trace.WriteLine("Lock enabled");
		}

		void connection_LockDisabledMessageReceived(Connection context, string message) {
			Locked = false;
			Trace.WriteLine("Lock disabled");
		}

		void connection_OutputChannelSelectMessageReceived(Connection context, int channel) {
			CurrentChannelIndex = channel - 1;
			Trace.WriteLine(String.Format("Output channel {0} selected", channel));
		}

		void connection_InputChannelSelectMessageReceived(Connection context, int channel) {
			ChangeInputChannel(channel);
		}

		public void ChangeInputChannel(int channel) {
			if (AllChannelsSelected) {
				Trace.WriteLine(String.Format("Changing all channels to input {0}", channel));
				for (int i = 0; i < 8; i++) {
					if (Mode == SwitchMode.Audio || Mode == SwitchMode.AudioVideo)
						AudioChannels[i] = channel;
					if (Mode == SwitchMode.Video || Mode == SwitchMode.AudioVideo)
						VideoChannels[i] = channel;
				}
			} else if (CurrentChannelIndex >= 0) {
				Trace.WriteLine(String.Format("Changing channel {0} to input {1}", CurrentChannelIndex + 1, channel));
				if (Mode == SwitchMode.Audio || Mode == SwitchMode.AudioVideo)
					AudioChannels[CurrentChannelIndex] = channel;
				if (Mode == SwitchMode.Video || Mode == SwitchMode.AudioVideo)
					VideoChannels[CurrentChannelIndex] = channel;
			}
			AllChannelsSelected = false;
			CurrentChannelIndex = -1;
		}

		void connection_MessageReceived(Connection context, string message) {
			//Trace.WriteLine(message);
		}

		void connection_AllOutputChannelsSelectedMessageReceived(Connection context, string message) {
			SelectAllOutputChannels();
		}

		public void SelectAllOutputChannels() {
			AllChannelsSelected = true;
			CurrentChannelIndex = -1;
			Trace.WriteLine("All output channels selected");
		}

		public void Update(string message) {
			if (message.StartsWith("0RA")) {
				int outputChannel = int.Parse(message.Substring(3, 1));
				int inputChannel = int.Parse(message.Substring(4, 1));

				AudioChannels[outputChannel - 1] = inputChannel;
			} else if (message.StartsWith("0RV")) {
				int outputChannel = int.Parse(message.Substring(3, 1));
				int inputChannel = int.Parse(message.Substring(4, 1));

				VideoChannels[outputChannel - 1] = inputChannel;
			} else if (message == "0RMAV") {
				Mode = SwitchMode.AudioVideo;
			} else if (message == "0RMA") {
				Mode = SwitchMode.Audio;
			} else if (message == "ORMV") {
				Mode = SwitchMode.Video;
			} else if (message == "0ILOF") {
				Locked = false;
			} else if (message == "0ILON") {
				Locked = true;
			}
		}
	}
}
